package SW02;

use Exporter;
@ISA = (Exporter);
@EXPORT = qw(sw);

use DataFile01;
    
sub sw{

# Set query and database sequences
my $d_id = shift;
my $q_seq = shift;
my $d_seq = shift;

# Gap and gap extension penalties
my $gap = -5;
my $ext = -1;

# Convert sequences into arrays
my @q_aas = split(//,$q_seq);
my @d_aas = split(//,$d_seq);

# Hash for conversion of amino acid to number
my $all_aa_str = "CSTPAGNDEQHRKMILVFYW";
my @all_aas = split(//,$all_aa_str);
my %an_cnvd;
my $num = 0;
foreach (@all_aas){
    $an_cnvd{$_} = $num;
    $num++;
}

# Convert amino acid alphapet to number
$_ = $an_cnvd{$_} foreach (@q_aas);
$_ = $an_cnvd{$_} foreach (@d_aas);

# 2D array for score matrix
my $ref_score_matrix = &return_blosum62;
my @score_matrix = @{$ref_score_matrix};

# Creat align matrix
my @align_matrix;
my $q_length = length($q_seq);
my $d_length = length($d_seq);

# Fill align matrix with "0"
for (0..$q_length){
    $align_matrix[$_] = [(0) x ($d_length+1)];
}

# Re-fill the first column/row with gap penalty
for (1..$q_length){
    $align_matrix[$_][0] = $_==1? $gap : $gap+($_-1)*$ext;
}
for (1..$d_length){
    $align_matrix[0][$_] = $_==1? $gap : $gap+($_-1)*$ext;
}

# Creat backtrack matrix
my @backtrack_matrix;

# Fill backtrack matrix with "0"
for (0..$q_length){
    $backtrack_matrix[$_] = [(0) x ($d_length+1)];
}

# Compute align/backtrack matrix
my $k = 0;
my $gap_value;
for (my $i=1; $i<=$q_length; $i++) {
    for (my $j=1; $j<=$d_length; $j++) {
        if ($k == 0){
            $gap_value = $gap;
        }else{
            $gap_value = $ext*$k;
        }
        my ($tmp_score, $k, $dir) =&max(
                $align_matrix[$i-1][$j-1]+$score_matrix[$q_aas[$i-1]][$d_aas[$j-1]],
                $align_matrix[$i][$j-1]+$gap_value,
                $align_matrix[$i-1][$j]+$gap_value,
                $k
        );
        $align_matrix[$i][$j] = $tmp_score;
        $backtrack_matrix[$i][$j] = $dir;
    }
}

# Subroutine for finding max score
sub max{
    my($d,$u,$l,$k)=@_;
    if($d>$u){
        return ($d, 0, 0) if($d>$l && $d>0);
        return (0, 0, 0) if($d>$l && $d<=0);
        return ($l, $k+1, 2) if($l>0);
        return (0, $k+1, 2) if($l<=0);
    }
    else{
        return ($u, $k+1, 1) if($u>$l && $u>0);
        return (0, $k+1, 1) if($u>$l && $u<0);
        return ($l, $k+1, 2) if($l>0);
        return (0, $k+1, 2) if($l<=0);  
    }
}

# Find max score and its row & col
my $max_score = 0;
my $r = 0;
my $c = 0;
my $max_row;
my $max_col;

foreach my $row (@align_matrix){
    my @array = @{$row};
    foreach my $col (@array){
        $max_score = $col if ($col >= $max_score);
        $max_row = $r if ($col >= $max_score);
        $max_col = $c if ($col >= $max_score);
        $c++;
    }
    $r++;
    $c = 0;
}

# Backtrack align_matrix
my $i = $max_row;
my $j = $max_col;
my @align1;
my @align2;

while ($i > 0 && $j > 0) {    
    # Score diagonal
    if ($backtrack_matrix[$i][$j] == 0) {
        push @align1, $q_aas[$i-1];
        push @align2, $d_aas[$j-1];
        $i--;
        $j--;
    }
    # Score up
    elsif ($backtrack_matrix[$i][$j] == 1){
        push @align1, "-";
        push @align2, $d_aas[$j-1];
        $j--;        
    }
    # Score left
    elsif ($backtrack_matrix[$i][$j] == 2) {
        push @align1, $q_aas[$i-1];
        push @align2, "-";
        $i--;
    }
    last if ($align_matrix[$i+1][$j+1]==0);
}

# Reverse
@align1 = reverse @align1;
@align2 = reverse @align2;

my @homol;
my $gap_num = 0;
my $match_num = 0;
my $sim_num = 0;
foreach (0..$#align1){
    if ($align1[$_] eq $align2[$_]){
        push @homol, '|';
        $match_num++;
    }elsif($align1[$_] eq '-' || $align2[$_] eq '-'){
        push @homol, ' ';
        $gap_num++;
    }elsif($align1[$_] ne $align2[$_]){
        my $score = $score_matrix[$align1[$_]][$align2[$_]];
        if ($score == 2 || $score == 3){
            push @homol, ':';
            $sim_num++;
        }elsif($score == 0 || $score == 1){
            push @homol, '.';
            $sim_num++;
        }else{
            push @homol, ' ';
        }
    }
}

# Convert number back to amino acid or "-"
foreach (@align1){
    $_ = $_ eq '-'?'-':$all_aas[$_];
}
foreach (@align2){
    $_ = $_ eq '-'?'-':$all_aas[$_];
}

# Print out result
my $M = length($q_seq);
my $N = length($d_seq);
my $S = $max_score;
my $bitS = ($S+11.749)/2.5932;
$bitS = sprintf("%.1f",$bitS);
my $K = 0.14;
my $LMD = 0.32;
my $tot_len = $#align1+1;
my $mtch_sim = $match_num+$sim_num;
my $Ipct = $match_num/$tot_len*100;
my $Ppct = $mtch_sim/$tot_len*100;
my $Gpct = $gap_num/$tot_len*100;
$Ipct = sprintf("%.1f",$Ipct);
$Ppct = sprintf("%.1f",$Ppct);
$Gpct = sprintf("%.1f",$Gpct);

my $Eval = $K*$M*$N*exp(-1*$LMD*$S);
$Eval = sprintf("%3e",$Eval);

print "[$d_id]\n";
print "Length = $M\n\n";
print "Score = $bitS bits ($S), Expect = $Eval\n";
print "Identities = $match_num/$tot_len ($Ipct%), Positives = $mtch_sim/$tot_len ($Ppct%), Gaps = $gap_num/$tot_len ($Gpct%)\n\n";
print "Query ";
print "$_"foreach @align1;
print "\n";
print "      ";
print "$_"foreach @homol;
print "\n";
print "Sbjct ";
print "$_"foreach @align2;
print "\n\n\n";

my $out_str = '';
$out_str .= "[$d_id]\n";
$out_str .= "Length = $M\n\n";
$out_str .= "Score = $bitS bits ($S), Expect = $Eval\n";
$out_str .= "Identities = $match_num/$tot_len ($Ipct%), Positives = $mtch_sim/$tot_len ($Ppct%), Gaps = $gap_num/$tot_len ($Gpct%)\n\n";
$out_str .= "Query ";
$out_str .= "$_"foreach @align1;
$out_str .= "\n";
$out_str .= "      ";
$out_str .= "$_"foreach @homol;
$out_str .= "\n";
$out_str .= "Sbjct ";
$out_str .= "$_"foreach @align2;
$out_str .= "\n\n\n";

return $out_str;

}

return 1;

